/****************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ****************************************************************************
     
     Module Name:
     ap_ids.c
     
     Abstract:
     monitor intrusion detection condition
     
     Revision History:
     Who         When          What
     --------    ----------    ----------------------------------------------
     Albert Yang 11-02-2007    created for RT2860
     John Li 12-28-2007           edit by RT61AP
     
 */

#include <net/iw_handler.h>
#include "rt_config.h"

#define IDS_EXEC_INTV          (1*HZ)       // 1 sec

#define PRINT_MAC(addr)	\
	addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]

// for wireless system event message
char const *pWirelessSysEventText[IW_SYS_EVENT_TYPE_NUM] = {    
	// system status event     
    "had associated successfully",							/* IW_ASSOC_EVENT_FLAG */
    "had disassociated",									/* IW_DISASSOC_EVENT_FLAG */
    "had deauthenticated",									/* IW_DEAUTH_EVENT_FLAG */
    "had been aged-out and disassociated",					/* IW_AGEOUT_EVENT_FLAG */
    "occurred CounterMeasures attack",						/* IW_COUNTER_MEASURES_EVENT_FLAG */	
    "occurred replay counter different in Key Handshaking",	/* IW_REPLAY_COUNTER_DIFF_EVENT_FLAG */
    "occurred RSNIE different in Key Handshaking",			/* IW_RSNIE_DIFF_EVENT_FLAG */
    "occurred MIC different in Key Handshaking",			/* IW_MIC_DIFF_EVENT_FLAG */
    "occurred ICV error in RX",								/* IW_ICV_ERROR_EVENT_FLAG */
    "occurred MIC error in RX",								/* IW_MIC_ERROR_EVENT_FLAG */
	"Group Key Handshaking timeout",						/* IW_GROUP_HS_TIMEOUT_EVENT_FLAG */ 
	"Pairwise Key Handshaking timeout",						/* IW_PAIRWISE_HS_TIMEOUT_EVENT_FLAG */ 
	"RSN IE sanity check failure",							/* IW_RSNIE_SANITY_FAIL_EVENT_FLAG */ 
	"set key done in WPA/WPAPSK",							/* IW_SET_KEY_DONE_WPA1_EVENT_FLAG */ 
	"set key done in WPA2/WPA2PSK",                         /* IW_SET_KEY_DONE_WPA2_EVENT_FLAG */ 
	"connects with our wireless client",                    /* IW_STA_LINKUP_EVENT_FLAG */ 
	"disconnects with our wireless client",                 /* IW_STA_LINKDOWN_EVENT_FLAG */
	"scan completed"										/* IW_SCAN_COMPLETED_EVENT_FLAG */
	};						

// for wireless IDS_spoof_attack event message
char const *pWirelessSpoofEventText[IW_SPOOF_EVENT_TYPE_NUM] = {   	
    "detected conflict SSID",								/* IW_CONFLICT_SSID_EVENT_FLAG */
    "detected spoofed association response",				/* IW_SPOOF_ASSOC_RESP_EVENT_FLAG */
    "detected spoofed reassociation responses",				/* IW_SPOOF_REASSOC_RESP_EVENT_FLAG */
    "detected spoofed probe response",						/* IW_SPOOF_PROBE_RESP_EVENT_FLAG */
    "detected spoofed beacon",								/* IW_SPOOF_BEACON_EVENT_FLAG */
    "detected spoofed disassociation",						/* IW_SPOOF_DISASSOC_EVENT_FLAG */
    "detected spoofed authentication",						/* IW_SPOOF_AUTH_EVENT_FLAG */
    "detected spoofed deauthentication",					/* IW_SPOOF_DEAUTH_EVENT_FLAG */
    "detected spoofed unknown management frame",			/* IW_SPOOF_UNKNOWN_MGMT_EVENT_FLAG */
	"detected replay attack"								/* IW_REPLAY_ATTACK_EVENT_FLAG */	
	};

// for wireless IDS_flooding_attack event message
char const *pWirelessFloodEventText[IW_FLOOD_EVENT_TYPE_NUM] = {   	
	"detected authentication flooding",						/* IW_FLOOD_AUTH_EVENT_FLAG */
    "detected association request flooding",				/* IW_FLOOD_ASSOC_REQ_EVENT_FLAG */
    "detected reassociation request flooding",				/* IW_FLOOD_REASSOC_REQ_EVENT_FLAG */
    "detected probe request flooding",						/* IW_FLOOD_PROBE_REQ_EVENT_FLAG */
    "detected disassociation flooding",						/* IW_FLOOD_DISASSOC_EVENT_FLAG */
    "detected deauthentication flooding",					/* IW_FLOOD_DEAUTH_EVENT_FLAG */
    "detected 802.1x eap-request flooding"					/* IW_FLOOD_EAP_REQ_EVENT_FLAG */	
	};

extern INT RTMPGetKeyParameter(
    IN  PCHAR   key,
    OUT PCHAR   dest,   
    IN  INT     destsize,
    IN  PCHAR   buffer);

VOID RTMPIdsStart(
	IN PRTMP_ADAPTER	pAd)
{

	if (pAd->PortCfg.IDSTimerRunning == FALSE)
	{		
		RTMPAddTimer(&pAd->PortCfg.IDSTimer, IDS_EXEC_INTV);
		pAd->PortCfg.IDSTimerRunning = TRUE;
	}
}

VOID RTMPIdsStop(
	IN PRTMP_ADAPTER	pAd)
{
	
	if (pAd->PortCfg.IDSTimerRunning == TRUE)
	{
		RTMPCancelTimer(&pAd->PortCfg.IDSTimer);
		pAd->PortCfg.IDSTimerRunning = FALSE;
	}
}

VOID RTMPHandleIdsEvent(
	IN PRTMP_ADAPTER	pAd)
{
	INT i, j;

	UINT32	FloodFrameCount[IW_FLOOD_EVENT_TYPE_NUM];
	UINT32	FloodFrameThreshold[IW_FLOOD_EVENT_TYPE_NUM];

	FloodFrameCount[0] = pAd->PortCfg.RcvdAuthCount;
	FloodFrameCount[1] = pAd->PortCfg.RcvdAssocReqCount;
	FloodFrameCount[2] = pAd->PortCfg.RcvdReassocReqCount;
	FloodFrameCount[3] = pAd->PortCfg.RcvdProbeReqCount;
	FloodFrameCount[4] = pAd->PortCfg.RcvdDisassocCount;
	FloodFrameCount[5] = pAd->PortCfg.RcvdDeauthCount;
	FloodFrameCount[6] = pAd->PortCfg.RcvdEapReqCount;

	FloodFrameThreshold[0] = pAd->PortCfg.AuthFloodThreshold;
	FloodFrameThreshold[1] = pAd->PortCfg.AssocReqFloodThreshold;
	FloodFrameThreshold[2] = pAd->PortCfg.ReassocReqFloodThreshold;
	FloodFrameThreshold[3] = pAd->PortCfg.ProbeReqFloodThreshold;
	FloodFrameThreshold[4] = pAd->PortCfg.DisassocFloodThreshold;
	FloodFrameThreshold[5] = pAd->PortCfg.DeauthFloodThreshold;
	FloodFrameThreshold[6] = pAd->PortCfg.EapReqFooldThreshold;

	// trigger flooding traffic event
	for (j = 0; j < IW_FLOOD_EVENT_TYPE_NUM; j++)
	{		
		if ((FloodFrameThreshold[j] > 0) && (FloodFrameCount[j] > FloodFrameThreshold[j]))
		{						
			RTMPSendWirelessEvent(pAd, IW_FLOOD_AUTH_EVENT_FLAG + j, NULL, MAX_MBSSID_NUM, 0);	
			//DBGPRINT(RT_DEBUG_TRACE, "flooding traffic event(%d) - %d\n", IW_FLOOD_AUTH_EVENT_FLAG + j, FloodFrameCount[j]);	
		}
	}	
	
	for (i = 0; i < pAd->PortCfg.BssidNum; i++)
	{	
		UINT32	SpoofedFrameCount[IW_SPOOF_EVENT_TYPE_NUM];
		CHAR	RssiOfSpoofedFrame[IW_SPOOF_EVENT_TYPE_NUM];
		INT		k;

		SpoofedFrameCount[0] = pAd->PortCfg.MBSSID[i].RcvdConflictSsidCount;
		SpoofedFrameCount[1] = pAd->PortCfg.MBSSID[i].RcvdSpoofedAssocRespCount;
		SpoofedFrameCount[2] = pAd->PortCfg.MBSSID[i].RcvdSpoofedReassocRespCount;
		SpoofedFrameCount[3] = pAd->PortCfg.MBSSID[i].RcvdSpoofedProbeRespCount;
		SpoofedFrameCount[4] = pAd->PortCfg.MBSSID[i].RcvdSpoofedBeaconCount;
		SpoofedFrameCount[5] = pAd->PortCfg.MBSSID[i].RcvdSpoofedDisassocCount;
		SpoofedFrameCount[6] = pAd->PortCfg.MBSSID[i].RcvdSpoofedAuthCount;
		SpoofedFrameCount[7] = pAd->PortCfg.MBSSID[i].RcvdSpoofedDeauthCount;
		SpoofedFrameCount[8] = pAd->PortCfg.MBSSID[i].RcvdSpoofedUnknownMgmtCount;
		SpoofedFrameCount[9] = pAd->PortCfg.MBSSID[i].RcvdReplayAttackCount;

		RssiOfSpoofedFrame[0] = pAd->PortCfg.MBSSID[i].RssiOfRcvdConflictSsid;
		RssiOfSpoofedFrame[1] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedAssocResp;
		RssiOfSpoofedFrame[2] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedReassocResp;
		RssiOfSpoofedFrame[3] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedProbeResp;
		RssiOfSpoofedFrame[4] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedBeacon;
		RssiOfSpoofedFrame[5] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedDisassoc;
		RssiOfSpoofedFrame[6] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedAuth;
		RssiOfSpoofedFrame[7] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedDeauth;
		RssiOfSpoofedFrame[8] = pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedUnknownMgmt;
		RssiOfSpoofedFrame[9] = pAd->PortCfg.MBSSID[i].RssiOfRcvdReplayAttack;

		// trigger spoofed attack event
		for (k = 0; k < IW_SPOOF_EVENT_TYPE_NUM; k++)
		{
			if (SpoofedFrameCount[k] > 0)
			{			
				RTMPSendWirelessEvent(pAd, IW_CONFLICT_SSID_EVENT_FLAG + k, NULL, i, RssiOfSpoofedFrame[k]);	
				//DBGPRINT(RT_DEBUG_TRACE, "spoofed attack event(%d) - %d\n", IW_CONFLICT_SSID_EVENT_FLAG + k, SpoofedFrameCount[k]);	
			}
		}					
	}

	
			
	
}

VOID RTMPClearAllIdsCounter(
	IN PRTMP_ADAPTER	pAd)
{
	INT	i;

	pAd->PortCfg.RcvdAuthCount = 0;
	pAd->PortCfg.RcvdAssocReqCount = 0;
	pAd->PortCfg.RcvdReassocReqCount = 0;
	pAd->PortCfg.RcvdProbeReqCount = 0;
	pAd->PortCfg.RcvdDisassocCount = 0;
	pAd->PortCfg.RcvdDeauthCount = 0;
	pAd->PortCfg.RcvdEapReqCount = 0;

	for (i = 0; i < pAd->PortCfg.BssidNum; i++)
	{
		pAd->PortCfg.MBSSID[i].RcvdConflictSsidCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedAssocRespCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedReassocRespCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedProbeRespCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedBeaconCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedDisassocCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedAuthCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedDeauthCount = 0;
		pAd->PortCfg.MBSSID[i].RcvdSpoofedUnknownMgmtCount = 0;	
		pAd->PortCfg.MBSSID[i].RcvdReplayAttackCount = 0;		

		pAd->PortCfg.MBSSID[i].RssiOfRcvdConflictSsid = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedAssocResp = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedReassocResp = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedProbeResp = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedBeacon = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedDisassoc = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedAuth = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedDeauth = 0;
		pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedUnknownMgmt = 0;	
		pAd->PortCfg.MBSSID[i].RssiOfRcvdReplayAttack = 0;		
		
	}
}

VOID RTMPIdsPeriodicExec(
    IN PRTMP_ADAPTER pAd)
{
	pAd->PortCfg.IDSTimerRunning = FALSE;

	// when IDS occured, send out wireless event
//	if (pAd->PortCfg.bWirelessEvent)	
		RTMPHandleIdsEvent(pAd);

	// clear all IDS counter
	RTMPClearAllIdsCounter(pAd);

	// set timer
	if (pAd->PortCfg.IdsEnable)
	{
		RTMPAddTimer(&pAd->PortCfg.IDSTimer, IDS_EXEC_INTV);
		pAd->PortCfg.IDSTimerRunning = TRUE;
	}	
}


/*
	========================================================================
	Routine Description:
		This routine is used to check if a rogue AP sent an 802.11 management
		frame to a client using our BSSID.
		
	Arguments:
		pAd    	- Pointer to our adapter
		pHeader - Pointer to 802.11 header

	Return Value:
		TRUE	- This is a spoofed frame
		FALSE	- This isn't a spoofed frame

	========================================================================
*/
BOOLEAN RTMPSpoofedMgmtDetection(
	IN PRTMP_ADAPTER	pAd,
	IN PHEADER_802_11 	pHeader,
	IN CHAR				Rssi)
{
	INT	i;	

	for (i = 0; i < pAd->PortCfg.BssidNum; i++)
	{
		// Spoofed BSSID detection
		if (NdisEqualMemory(pHeader->Addr2, pAd->PortCfg.MBSSID[i].Bssid, MAC_ADDR_LEN))
		{
			CHAR RcvdRssi;
		
			RcvdRssi = ConvertToRssi(pAd, Rssi, RSSI_NO_1);
		
			switch (pHeader->FC.SubType) 
			{
				case SUBTYPE_ASSOC_RSP:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedAssocRespCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedAssocResp = RcvdRssi;
					break;

				case SUBTYPE_REASSOC_RSP:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedReassocRespCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedReassocResp = RcvdRssi;
					break;

				case SUBTYPE_PROBE_RSP:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedProbeRespCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedProbeResp = RcvdRssi;
					break;

				case SUBTYPE_BEACON:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedBeaconCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedBeacon = RcvdRssi;
					break;	

				case SUBTYPE_DISASSOC:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedDisassocCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedDisassoc = RcvdRssi;
					break;
					
				case SUBTYPE_AUTH:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedAuthCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedAuth = RcvdRssi;
					break;

				case SUBTYPE_DEAUTH:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedDeauthCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedDeauth = RcvdRssi;
					break;

				default:
					pAd->PortCfg.MBSSID[i].RcvdSpoofedUnknownMgmtCount ++;
					pAd->PortCfg.MBSSID[i].RssiOfRcvdSpoofedUnknownMgmt = RcvdRssi;
					break;	

			}					

			return TRUE;
			
		}
		
	}

	return FALSE;
}


VOID RTMPConflictSsidDetection(
	IN PRTMP_ADAPTER	pAd,
	IN PUCHAR			pSsid,
	IN UCHAR			SsidLen,
	IN CHAR				Rssi)
{
	INT	i;
	
	for (i = 0; i < pAd->PortCfg.BssidNum; i++)
	{
		// Conflict SSID detection
		if (SSID_EQUAL(pSsid, SsidLen, pAd->PortCfg.MBSSID[i].Ssid, pAd->PortCfg.MBSSID[i].SsidLen))
		{
			CHAR RcvdRssi;
		
			RcvdRssi = ConvertToRssi(pAd, Rssi, RSSI_NO_1);

			pAd->PortCfg.MBSSID[i].RcvdConflictSsidCount ++;
			pAd->PortCfg.MBSSID[i].RssiOfRcvdConflictSsid = RcvdRssi;
			return;
		}	
	}			
}

BOOLEAN RTMPReplayAttackDetection(
	IN PRTMP_ADAPTER	pAd,
	IN PUCHAR			pAddr2,
	IN CHAR				Rssi)
{
	INT	i;
	
	for (i = 0; i < pAd->PortCfg.BssidNum; i++)
	{
		// Conflict BSSID detection
		if (NdisEqualMemory(pAddr2, pAd->PortCfg.MBSSID[i].Bssid, MAC_ADDR_LEN))
		{
			CHAR RcvdRssi;
		
			RcvdRssi = ConvertToRssi(pAd, Rssi, RSSI_NO_1);
		
			pAd->PortCfg.MBSSID[i].RcvdReplayAttackCount ++;
			pAd->PortCfg.MBSSID[i].RssiOfRcvdReplayAttack = RcvdRssi;
			return TRUE;
		}	
	}	
	return FALSE;
}

VOID RTMPUpdateStaMgmtCounter(
	IN PRTMP_ADAPTER	pAd,
	IN USHORT			type)
{
		
	switch (type)
	{
		case SUBTYPE_ASSOC_REQ:
          	pAd->PortCfg.RcvdAssocReqCount ++;
			//printk("RcvdAssocReqCount=%d\n", pAd->PortCfg.RcvdAssocReqCount);
			break;

		case SUBTYPE_REASSOC_REQ:
          	pAd->PortCfg.RcvdReassocReqCount ++;
			//printk("RcvdReassocReqCount=%d\n", pAd->PortCfg.RcvdReassocReqCount);
			break;

		case SUBTYPE_PROBE_REQ:
          	pAd->PortCfg.RcvdProbeReqCount ++;
			//printk("RcvdProbeReqCount=%d\n", pAd->PortCfg.RcvdProbeReqCount);
			break;	

		case SUBTYPE_DISASSOC:
          	pAd->PortCfg.RcvdDisassocCount ++;
			//printk("RcvdDisassocCount=%d\n", pAd->PortCfg.RcvdDisassocCount);
			break;

		case SUBTYPE_DEAUTH:
          	pAd->PortCfg.RcvdDeauthCount ++;
			//printk("RcvdDeauthCount=%d\n", pAd->PortCfg.RcvdDeauthCount);
			break;

		case SUBTYPE_AUTH:
          	pAd->PortCfg.RcvdAuthCount ++;
			//printk("RcvdAuthCount=%d\n", pAd->PortCfg.RcvdAuthCount);
			break;	

	}
	
}

VOID rtmp_read_ids_from_file(
			IN  PRTMP_ADAPTER pAd,
			char *tmpbuf,
			char *buffer)
{	
	//IdsEnable
	if(RTMPGetKeyParameter("IdsEnable", tmpbuf, 10, buffer))
	{						
		if (simple_strtol(tmpbuf, 0, 10) == 1)
			pAd->PortCfg.IdsEnable = TRUE;
		else
			pAd->PortCfg.IdsEnable = FALSE;
						
		DBGPRINT(RT_DEBUG_TRACE, "IDS is %s\n", pAd->PortCfg.IdsEnable ? "enabled" : "disabled");
	}

	//AuthFloodThreshold
	if(RTMPGetKeyParameter("AuthFloodThreshold", tmpbuf, 10, buffer))
	{						
		pAd->PortCfg.AuthFloodThreshold = simple_strtol(tmpbuf, 0, 10);
							
		DBGPRINT(RT_DEBUG_TRACE, "AuthFloodThreshold = %d\n", pAd->PortCfg.AuthFloodThreshold);
	}

	//AssocReqFloodThreshold
	if(RTMPGetKeyParameter("AssocReqFloodThreshold", tmpbuf, 10, buffer))
	{						
		pAd->PortCfg.AssocReqFloodThreshold = simple_strtol(tmpbuf, 0, 10);
							
		DBGPRINT(RT_DEBUG_TRACE, "AssocReqFloodThreshold = %d\n", pAd->PortCfg.AssocReqFloodThreshold);
	}

	//ReassocReqFloodThreshold
	if(RTMPGetKeyParameter("ReassocReqFloodThreshold", tmpbuf, 10, buffer))
	{						
		pAd->PortCfg.ReassocReqFloodThreshold = simple_strtol(tmpbuf, 0, 10);
							
		DBGPRINT(RT_DEBUG_TRACE, "ReassocReqFloodThreshold = %d\n", pAd->PortCfg.ReassocReqFloodThreshold);
	}

	//ProbeReqFloodThreshold
	if(RTMPGetKeyParameter("ProbeReqFloodThreshold", tmpbuf, 10, buffer))
	{						
		pAd->PortCfg.ProbeReqFloodThreshold = simple_strtol(tmpbuf, 0, 10);
							
		DBGPRINT(RT_DEBUG_TRACE, "ProbeReqFloodThreshold = %d\n", pAd->PortCfg.ProbeReqFloodThreshold);
	}

	//DisassocFloodThreshold
	if(RTMPGetKeyParameter("DisassocFloodThreshold", tmpbuf, 10, buffer))
	{						
		pAd->PortCfg.DisassocFloodThreshold = simple_strtol(tmpbuf, 0, 10);
							
		DBGPRINT(RT_DEBUG_TRACE, "DisassocFloodThreshold = %d\n", pAd->PortCfg.DisassocFloodThreshold);
	}

	//DeauthFloodThreshold
	if(RTMPGetKeyParameter("DeauthFloodThreshold", tmpbuf, 10, buffer))
	{						
		pAd->PortCfg.DeauthFloodThreshold = simple_strtol(tmpbuf, 0, 10);
							
		DBGPRINT(RT_DEBUG_TRACE, "DeauthFloodThreshold = %d\n", pAd->PortCfg.DeauthFloodThreshold);
	}

	//EapReqFooldThreshold
	if(RTMPGetKeyParameter("EapReqFooldThreshold", tmpbuf, 10, buffer))
	{						
		pAd->PortCfg.EapReqFooldThreshold = simple_strtol(tmpbuf, 0, 10);
							
		DBGPRINT(RT_DEBUG_TRACE, "EapReqFooldThreshold = %d\n", pAd->PortCfg.EapReqFooldThreshold);
	}
}


/*
	========================================================================
	
	Routine Description:
		Send log message through wireless event

		Support standard iw_event with IWEVCUSTOM. It is used below.

		iwreq_data.data.flags is used to store event_flag that is defined by user. 
		iwreq_data.data.length is the length of the event log.

		The format of the event log is composed of the entry's MAC address and
		the desired log message (refer to pWirelessEventText).

			ex: 11:22:33:44:55:66 has associated successfully

		p.s. The requirement of Wireless Extension is v15 or newer. 

	========================================================================
*/
VOID RTMPSendWirelessEvent(
	IN	PRTMP_ADAPTER	pAd,
	IN	USHORT			Event_flag,
	IN	PUCHAR 			pAddr,
	IN	UCHAR			BssIdx,
	IN	CHAR			Rssi)
{
#if WIRELESS_EXT >= 15

	union 	iwreq_data      wrqu;
	PUCHAR 	pBuf = NULL, pBufPtr = NULL;
	USHORT	event, type, BufLen;	
	UCHAR	event_table_len = 0;

	type = Event_flag & 0xFF00;	
	event = Event_flag & 0x00FF;

	switch (type)
	{
		case IW_SYS_EVENT_FLAG_START:
			event_table_len = IW_SYS_EVENT_TYPE_NUM;
			break;

		case IW_SPOOF_EVENT_FLAG_START:
			event_table_len = IW_SPOOF_EVENT_TYPE_NUM;
			break;

		case IW_FLOOD_EVENT_FLAG_START:
			event_table_len = IW_FLOOD_EVENT_TYPE_NUM;
			break;
	}
	
	if (event_table_len == 0)
	{
		DBGPRINT(RT_DEBUG_ERROR, "%s : The type(%0x02x) is not valid.\n", __FUNCTION__, type);			       		       		
		return;
	}
	
	if (event >= event_table_len)
	{
		DBGPRINT(RT_DEBUG_ERROR, "%s : The event(%0x02x) is not valid.\n", __FUNCTION__, event);			       		       		
		return;
	}	
 
	//Allocate memory and copy the msg.
	if((pBuf = kmalloc(IW_CUSTOM_MAX_LEN, GFP_ATOMIC)) != NULL)
	{
		//Prepare the payload 
		memset(pBuf, 0, IW_CUSTOM_MAX_LEN);

		pBufPtr = pBuf;		

		if (pAddr)
			pBufPtr += sprintf(pBufPtr, "(RT61AP) STA(%02x:%02x:%02x:%02x:%02x:%02x) ", PRINT_MAC(pAddr));				
		else if (BssIdx < MAX_MBSSID_NUM)
			pBufPtr += sprintf(pBufPtr, "(RT61AP) BSS(ra%d) ", BssIdx);
		else
			pBufPtr += sprintf(pBufPtr, "(RT61AP) ");

		if (type == IW_SYS_EVENT_FLAG_START)
			pBufPtr += sprintf(pBufPtr, "%s", pWirelessSysEventText[event]);
		else if (type == IW_SPOOF_EVENT_FLAG_START)
			pBufPtr += sprintf(pBufPtr, "%s (RSSI=%d)", pWirelessSpoofEventText[event], Rssi);
		else if (type == IW_FLOOD_EVENT_FLAG_START)
			pBufPtr += sprintf(pBufPtr, "%s", pWirelessFloodEventText[event]);
		else
			pBufPtr += sprintf(pBufPtr, "%s", "unknown event");
		
		pBufPtr[pBufPtr - pBuf] = '\0';
		BufLen = pBufPtr - pBuf;
		
		memset(&wrqu, 0, sizeof(wrqu));	
		wrqu.data.flags = Event_flag;
		wrqu.data.length = BufLen;	
		
		//send wireless event
		wireless_send_event(pAd->net_dev, IWEVCUSTOM, &wrqu, pBuf);
	
		//DBGPRINT(RT_DEBUG_TRACE, "%s : %s\n", __FUNCTION__, pBuf);	
	
		kfree(pBuf);
	}
	else
		DBGPRINT(RT_DEBUG_ERROR, "%s : Can't allocate memory for wireless event.\n", __FUNCTION__);			       		       				
#else
	DBGPRINT(RT_DEBUG_ERROR, "%s : The Wireless Extension MUST be v15 or newer.\n", __FUNCTION__);	
#endif  /* WIRELESS_EXT >= 15 */  
}

